#!/usr/bin/env python
# -*- coding: utf-8 -*-
import string
import os
import sqlite3


class Util:

    def __init__(self):
        # Stopwords alphabetisch sortiert
        a = ["aber", "als", "am", "an", "auch", "auf", "aus"]
        b = ["bei", "bin", "bis", "bist"]
        c = []
        d = ["da", "dadurch", "daher", "darum", "das", "daß", "dass", "dein", "deine", "dem", "den", "der",
             "des", "dessen", "deshalb", "die", "dies", "dieser", "dieses", "doch", "dort", "du", "durch"]
        e = ["ein", "eine", "einem", "einen", "einer",
             "eines", "er", "es", "euer", "eure"]
        f = ["für"]
        g = []
        h = ["hatte", "hatten", "hattest", "hattet", "hier", "hinter"]
        i = ["ich", "ihr", "ihre", "im", "in", "ist"]
        j = ["ja", "jede", "jedem", "jeden", "jeder",
             "jedes", "jener", "jenes", "jetzt"]
        k = ["kann", "kannst", "können", "könnt"]
        l = []
        m = ["machen", "man", "mein", "meine", "mit",
             "muß", "mußt", "musst", "müssen", "müßt"]
        n = ["nach", "nachdem", "nein", "nicht", "nun"]
        o = ["oder"]
        p = []
        q = []
        r = []
        s = ["seid", "sein", "seine", "sich", "sie", "sind", "soll",
             "sollen", "sollst", "sollt", "sonst", "soweit", "sowie"]
        t = []
        u = ["über", "und", "unser", "unsere", "unter"]
        v = ["vom", "von", "vor"]
        w = ["wann", "warum", "was", "weiter", "weitere", "wenn", "wer", "werde", "werden", "werdet",
             "weshalb", "wie", "wieder", "wieso", "wir", "wird", "wirst", "wo", "woher", "wohin"]
        x = []
        y = []
        z = ["zu", "zum", "zur"]

        # Liste aller Stopwords
        self.__stopwords = a + b + c + d + e + f + g + h + i + j + k + \
            l + m + n + o + p + q + r + s + t + u + v + w + x + y + z

    # Extrahiert Schlüsselwörter aus einer Frage
    def extractKeywords(self, question):
        # Alle Woerter der Frage
        words = question.split()
        # Woerter kleingeschrieben
        lower_words = [word.lower() for word in words]
        # Satzzeichen entfernt
        clean_words = [word.translate(None, string.punctuation)
                       for word in lower_words]
        # Stopwoerter entfernt
        keywords = [
            word for word in clean_words if word not in self.__stopwords]
        return keywords

    # Liefert die Potenzmenge von s abzüglich der leeren Menge
    def power_set_less_empty_set(self, s):

        power_set = [set()]

        for element in s:
            one_element_set = {element}
            power_set += [subset | one_element_set for subset in power_set]

        power_set.remove(set([]))

        return power_set


class Chatbot:

    def __init__(self):
        self.util = Util()
        self.conversation = True

    # Startet die Konversation mit dem Nutzer
    def conversate(self):
        # Begruessung
        print "Hallo, ich bin der DKRZ-Chatbot."

        # Bis der Nutzer die Konversation beendet
        while self.conversation:
            question = raw_input("Frag mich, was du wissen willst.\n")
            keywords = self.util.extractKeywords(question)
            links = self.linksFromDB(keywords)
            if links is None:
                print "Ueber dieses Thema weiss ich leider nichts."
            elif len(links) == 0:
                print "Ueber dieses Thema weiss ich leider nichts."
            else:
                print "Zu diesem Thema koennten folgende Webseiten hilfreich sein:"
                for link in links:
                    print link

            # Ueberprueft, ob die Konversation beendet werden soll und reagiert
            # entsprechend
            while True:
                end_of_conversation = raw_input(
                    "Möchtest du die Konversation beenden? (Schreibe J oder N)\n")
                end_of_conversation = end_of_conversation.lower()
                if end_of_conversation == "j":
                    self.conversation = False
                    break
                elif end_of_conversation == "n":
                    self.conversation = True
                    break
                else:
                    print "Das habe ich nicht verstanden."

    # Liefert eine Menge von Links aus der Datenbank zu einem Iterable von
    # Schlüsselwörtern
    def linksFromDB(self, keywords):

        connection = sqlite3.connect("words.db")
        cursor = connection.cursor()
        connection.text_factory = str

        # Beinhaltet die Linkmengen aller Schluesselwoerter
        all_wordlinks = []

        for keyword in keywords:
            z = 1  # dummy statement
            sqlCommand = 'SELECT link FROM wordlinks WHERE word="' + keyword +'"' 
            cursor.execute(sqlCommand)
            keyword_links = cursor.fetchall()
            all_wordlinks.append(tuple(keyword_links))        

        # Potenzmenge der Linkmengen
        power_set_wordlinks = self.util.power_set_less_empty_set(all_wordlinks)

        # Erstellt eine Menge von Links, die zu der groestmoeglichen Anzahl von
        # Links gehoert und gibt diese aus
        for x in range(len(keywords), 0, -1):

            # Menge aller Linkmengen mit Laenge x
            size_x_wordlinks = []

            for power_set_subset in power_set_wordlinks:
                if len(power_set_subset) == x:
                    size_x_wordlinks.append(power_set_subset)

            # Schnittmenge der Linkmengen mit Laenge x
            intersection = []

            for size_x_subset in size_x_wordlinks:
                # Schnittmenge startet mit allen Links zu einem Wort
                intersection = list(next(iter(size_x_subset)))

                # Es wird der Schnitt mit den Links jedes anderen Wortes
                # gebildet
                for wordlinks in size_x_subset:
                    intersection = [
                        wordlink for wordlink in intersection if wordlink in wordlinks]

                # Wenn der Schnitt nicht leer ist, wurden Links mit
                # groestmoeglicher Wortzugehoerigkeit gefunden und der Schnitt wird
                # ausgegeben
                if len(intersection) > 0:
                    return intersection

        # Wenn zu den Schluesselwoertern keine Links in der Datenbank vorhanden
        # sind, wird eine leere Liste ausgegeben
        return []


chatbot = Chatbot()
chatbot.conversate()
